#include <dlfcn.h>
#include <string>
#include <vector>
#include <cctype>
#include <map>
#include <set>

#include <sys/types.h>
#include <arpa/telnet.h>
#include <sys/socket.h>

#include "coroutine.h"
#include "colors.hh"
#include "setfun.hh"
#include "strfun.hh"
#include "tagfun.hh"
#include "fmtfun.hh"

static struct LibcGet
{
    void *f;
    LibcGet() : f(dlopen("libc.so.6",RTLD_NOW)) {}
    ~LibcGet() { dlclose(f); }
    void *operator() (const char *s) { return dlsym(f, s); }  
} caller;

static bool is_valid_tagname_char(const char c)
{
    return  (c >= 'A' && c <= 'Z')
         || (c >= 'a' && c <= 'z')
         || (c >= '0' && c <= '9')
         || (c == '-' || c == '_' || c == '!');
}
static bool is_valid_unquoted_tagattr_char(const char c)
{
    return is_valid_tagname_char(c)
        || c=='-' || c=='+'
        || c=='%' || c=='#'
        || c=='*';
}

class Transceiver
{
private:
    std::string PendingResult;
    std::string PendingFeedback;
    std::string PendingMacro;
    enum {
        DefaultMode,
        SecureMode,
        LockedMode,
        OpenMode
    } CurrentMode;
    enum {
        ModeLocked,
        UntilNl,
        UntilTagEnd
    } ModeMode;
    bool ClosingTagsMode;
    std::vector<std::string> LineEndCloseTags;
    
    struct Formatting Formatting;
    
    std::map<std::string/*tagname*/, TagDef> TagDefs;
    std::map<std::string, EntityDef> EntityDefs;
    
private:
    void DefaultTagHandler(const TagDef& def, const Tag& tag)
    {
        // This should just do the macro expansion.
        // However, if this is a closing tag, it should also
        // do the closing versions of the tags included in the macro.
        // For this, we use the esc[2;0z and esc[2;1z macroes.
        
        std::string macro = def.MacroExpansion;
        if(!macro.empty())
        {
            if(tag.IsClosingTag)
                macro = "\33[2;1z" + macro + "\33[2;0z";
            DoMacroExpansion(macro);
        }
    }

    void HandleElementMetaTag(const TagDef& def, const Tag& tag)
    {
        AttrMatchResult attrs = DoAttrMatching(def, tag);
        std::string elname = UCase(attrs.matched["ELEMENT-NAME"]);
        {std::map<std::string, TagDef>::iterator i = TagDefs.find(elname);
        if(i != TagDefs.end())
        {
            /* Check if we can redefine the tag */
            if(!i->second.Is_OpenTag && CurrentMode != SecureMode) return; // ignore!
        }}
        TagDef& result = TagDefs[elname];
        result.Handler = &Transceiver::DefaultTagHandler;
        result.MacroExpansion = attrs.matched["DEFINITION"];
        result.Tag            = std::atoi(attrs.matched["TAG"].c_str());
        result.Flags          = attrs.matched["FLAG"];
        result.AttrList       = ParseAttrList(attrs.matched["ATT"], true);
        result.Is_EmptyTag = false;
        result.Is_OpenTag  = false;
        for(unsigned a=0; a<attrs.unmatched.size(); ++a)
        {
            const std::string param = UCase(attrs.unmatched[a]);
            if(param == "OPEN") result.Is_OpenTag = true;
            if(param == "EMPTY") result.Is_EmptyTag = true;
            if(param == "DELETE") { TagDefs.erase(elname); return; }
        }
    }
    void HandleAttListMetaTag(const TagDef& def, const Tag& tag)
    {
        AttrMatchResult attrs = DoAttrMatching(def, tag);
        std::string elname = UCase(attrs.matched["ELEMENT-NAME"]);
        std::map<std::string, TagDef>::iterator i = TagDefs.find(elname);
        if(i == TagDefs.end()) return;
        /* Check if we can redefine the tag */
        if(!i->second.Is_OpenTag && CurrentMode != SecureMode) return; // ignore!
        i->second.AttrList = ParseAttrList(attrs.matched["ATTRIBUTE-LIST"], true);
    }
    void HandleEntityMetaTag(const TagDef& def, const Tag& tag)
    {
        AttrMatchResult attrs = DoAttrMatching(def, tag);
        std::string enname = attrs.matched["NAME"];
        {std::map<std::string, EntityDef>::iterator i = EntityDefs.find(enname);
        if(i != EntityDefs.end())
        {
            /* Check if we can redefine the tag */
            if(!i->second.Is_OpenEnt && CurrentMode != SecureMode) return; // ignore!
        }}
        EntityDef& result = EntityDefs[enname];

        const std::string& newvalue = attrs.matched["VALUE"];
        
        enum { Replace, Delete, Add, Remove } mode = Replace;
        result.Desc  = attrs.matched["DESC"];
        result.Is_Recursive = true;
        result.Is_OpenEnt = true;
        for(unsigned a=0; a<attrs.unmatched.size(); ++a)
        {
            const std::string param = UCase(attrs.unmatched[a]);
            if(param == "PRIVATE") result.Is_PubEnt = result.Is_OpenEnt = false;
            if(param == "PUBLISH") result.Is_PubEnt = true;
            if(param == "DELETE") mode = Delete;
            if(param == "ADD")    mode = Add;
            if(param == "REMOVE") mode = Remove;
        }
        switch(mode)
        {
            case Replace: result.Value = newvalue; break;
            case Delete:  { EntityDefs.erase(enname); return; }
            case Add: case Remove:
                std::set<std::string> curvalues = ExplodeSet("|", result.Value);
                std::set<std::string> newvalues = ExplodeSet("|", newvalue);
                // remove.
                for(std::set<std::string>::const_iterator
                    i = newvalues.begin(); i != newvalues.end(); ++i)
                        curvalues.erase(*i);
                
                // remove + add.
                if(mode == Add)
                {
                    for(std::set<std::string>::const_iterator
                        i = newvalues.begin(); i != newvalues.end(); ++i)
                            curvalues.insert(*i);
                }
                result.Value = JoinSet("|", curvalues);
                break;
        }
    }
    void HandleTagMetaTag(const TagDef& def, const Tag& tag)
    {
        AttrMatchResult attrs = DoAttrMatching(def, tag);
        
        const int TagValue = atoi(attrs.matched["INDEX"].c_str());
        
        const std::string AutoTagName = "$AUTOTAG_" + attrs.matched["INDEX"];
        bool was_automatic = true;
        std::string elname = AutoTagName;
        
        for(std::map<std::string/*tagname*/, TagDef>::const_iterator
            i = TagDefs.begin(); i != TagDefs.end(); ++i)
        {
            if(i->second.Tag == TagValue)
            {
                /* Touch a predefined element instead */
                was_automatic = false;
                elname = i->first;
                break;
            }
        }
        
        
        TagDef& result = TagDefs[elname];
        
        std::string SubstituteTag = "<FONT";
        { const std::string& fore = attrs.matched["FORE"];
          if(!fore.empty()) SubstituteTag += " COLOR='" + fore + "'"; }
        { const std::string& back = attrs.matched["BACK"];
          if(!back.empty()) SubstituteTag += " BACK='" + back + "'"; }
        SubstituteTag += ">";
        
        // Window name is not supported.
        // Gag is not supported.
        
        result.Is_EmptyTag    = false;
        result.Tag            = TagValue;
        result.MacroExpansion = SubstituteTag;
        result.Handler        = &Transceiver::DefaultTagHandler;
        for(unsigned a=0; a<attrs.unmatched.size(); ++a)
        {
            const std::string param = UCase(attrs.unmatched[a]);
            if(param == "GAG") { /* not supported */ }
            if(param == "ENABLE") { /* implicit */ }
            if(param == "DISABLE")
            {
                if(was_automatic) { TagDefs.erase(elname); return; }
                result.MacroExpansion.clear();
            }
        }
    }
    
    void HandleBoldTag(const TagDef& def, const Tag& tag)
        { Formatting.Add(Formatting.B, tag.IsClosingTag?-1:1); RedoFormatting(); }
    void HandleItalicTag(const TagDef& def, const Tag& tag)
        { Formatting.Add(Formatting.I, tag.IsClosingTag?-1:1); RedoFormatting(); }
    void HandleUnderlineTag(const TagDef& def, const Tag& tag)
        { Formatting.Add(Formatting.U, tag.IsClosingTag?-1:1); RedoFormatting(); }
    void HandleStrikeoutTag(const TagDef& def, const Tag& tag)
        { Formatting.Add(Formatting.S, tag.IsClosingTag?-1:1); RedoFormatting(); }
    void HandleColorTag(const TagDef& def, const Tag& tag)
    {
        if(tag.IsClosingTag) { Formatting.PopFontDef(); RedoFormatting(); return; }
        FontDef fnt;
        AttrMatchResult attrs = DoAttrMatching(def, tag);
        fnt.forecolordefs = attrs.matched["FORE"];
        fnt.backcolordefs = attrs.matched["BACK"];
        Formatting.PushFontDef(fnt);
        RedoFormatting();
    }
    void HandleBrightTag(const TagDef& def, const Tag& tag)
        { Formatting.Add(Formatting.Bright, tag.IsClosingTag?-1:1); RedoFormatting(); }
    void HandleFontTag(const TagDef& def, const Tag& tag)
    {
        if(tag.IsClosingTag) { Formatting.PopFontDef(); RedoFormatting(); return; }
        FontDef fnt;
        AttrMatchResult attrs = DoAttrMatching(def, tag);
        fnt.fontname = attrs.matched["FACE"];
        fnt.fontsize = attrs.matched["SIZE"];
        fnt.forecolordefs = attrs.matched["COLOR"];
        fnt.backcolordefs = attrs.matched["BACK"];
        Formatting.PushFontDef(fnt);
        RedoFormatting();
    }
    
    void HandleNobrTag(const TagDef& def, const Tag& tag) { }
    void HandleParaTag(const TagDef& def, const Tag& tag) { }
    void HandleBRTag(const TagDef& def, const Tag& tag) { }
    void HandleSBRTag(const TagDef& def, const Tag& tag) { }

    void HandleVersionTag(const TagDef& def, const Tag& tag)
    {
        std::string VersionReply =
            "<VERSION MXP='1.0' CLIENT='bat_mxp' VERSION='" VERSION "'>";
        DoFeedback(VersionReply);
    }
    void HandleSupportTag(const TagDef& def, const Tag& tag)
    {
        std::set<std::string> responses;
        std::map<std::string, std::vector<std::string> > queries;
        AttrMatchResult attrs = DoAttrMatching(def, tag);
        for(unsigned a=0; a<attrs.unmatched.size(); ++a)
        {
            const std::string& token = attrs.unmatched[a];
            std::string::size_type dotpos = token.find('.');
            if(dotpos == token.npos)
                queries[token];
            else
                queries[token.substr(0, dotpos)].push_back(token.substr(dotpos+1));
        }
        if(queries.empty())
        {
            // Add an implicit query of all tags we have
            for(std::map<std::string/*tagname*/, TagDef>::const_iterator
                j = TagDefs.begin(); j != TagDefs.end(); ++j)
                queries[j->first];
        }
        for(std::map<std::string, std::vector<std::string> >::const_iterator
            i = queries.begin(); i != queries.end(); ++i)
        {
            std::map<std::string/*tagname*/, TagDef>::const_iterator
                j = TagDefs.find(UCase(i->first));
            if(j == TagDefs.end())
                responses.insert("-" + i->first);
            else
            {
                std::vector<std::string> subq = i->second; // make a copy for modifying
                if(subq.empty())
                    responses.insert("+" + i->first);
                else
                {
                    if(subq.size() == 1 && subq[0] == "*")
                    {
                        subq.clear();
                        for(unsigned b=0; b<j->second.AttrList.size(); ++b)
                            subq.push_back(j->second.AttrList[b].first);
                    }
                    /* Check each parameter */
                    for(unsigned a=0; a<subq.size(); ++a)
                    {
                        bool found=false;
                        for(unsigned b=0; b<j->second.AttrList.size(); ++b)
                            if(UCase(j->second.AttrList[b].first) == UCase(subq[a]))
                                { found=true; break; }
                        responses.insert((found?"+":"-") + i->first + "." + subq[a]);
                    }
                }
            }
        }
        /* Send response */
        std::string result = "<SUPPORTS";
        for(std::set<std::string>::const_iterator
            k = responses.begin(); k != responses.end(); ++k)
        {
            result += ' ';
            result += *k;
        }
        result += ">";
        DoFeedback(result);
    }

    void InitDefaultTags()
    {
        #define DefAt(t, name, defval) t.AttrList.push_back(std::make_pair(name,defval))
        
        /* COMMAND TAGS */
    
        {TagDef ElementMetaTag;
            DefAt(ElementMetaTag, "ELEMENT-NAME",""); // Not in specs
            DefAt(ElementMetaTag, "DEFINITION",  ""); // Not in specs
            DefAt(ElementMetaTag, "ATT",         "");
            DefAt(ElementMetaTag, "TAG",         "");
            DefAt(ElementMetaTag, "FLAG",        "");
            ElementMetaTag.Handler = &Transceiver::HandleElementMetaTag;
        TagDefs["!ELEMENT"] = TagDefs["!EL"]  = ElementMetaTag;}
        {TagDef AttListMetaTag;
            DefAt(AttListMetaTag, "ELEMENT-NAME", "");   // Not in specs
            DefAt(AttListMetaTag, "ATTRIBUTE-LIST", ""); // Not in specs
            AttListMetaTag.Handler = &Transceiver::HandleAttListMetaTag;
        TagDefs["!ATTLIST"] = TagDefs["!AT"] = AttListMetaTag;}
        {TagDef EntityMetaTag;
            DefAt(EntityMetaTag, "NAME", "");  // Not in specs
            DefAt(EntityMetaTag, "VALUE", ""); // Not in specs
            DefAt(EntityMetaTag, "DESC", "");
            EntityMetaTag.Handler = &Transceiver::HandleEntityMetaTag;
        TagDefs["!ENTITY"] = TagDefs["!EN"] = EntityMetaTag;}
        {TagDef TagMetaTag;
            DefAt(TagMetaTag, "INDEX", "");  // Not in specs
            DefAt(TagMetaTag, "WINDOWNAME", "");
            DefAt(TagMetaTag, "FORE", "");
            DefAt(TagMetaTag, "BACK", "");
            TagMetaTag.Handler = &Transceiver::HandleTagMetaTag;
        TagDefs["!TAG"] = TagMetaTag;}
        /* VAR not supported yet */
        
        /* ENTITIES */
        
        static const struct { const char* name; char chval; } entlist[] = {
         #include "entities.inc"
        };
        const unsigned n_entities = sizeof(entlist)/sizeof(*entlist);
        for(unsigned a=0; a<n_entities; ++a)
            { EntityDef ent; ent.Value += entlist[a].chval;
              EntityDefs[entlist[a].name]=ent; }
        
        /* TEXT FORMATTING TAGS (OPEN) */
        {TagDef Tag; Tag.Is_EmptyTag = false; Tag.Is_OpenTag = true;
            Tag.Handler = &Transceiver::HandleBoldTag;
         TagDefs["B"] = TagDefs["STRONG"] = TagDefs["BOLD"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = false; Tag.Is_OpenTag = true;
            Tag.Handler = &Transceiver::HandleItalicTag;
         TagDefs["I"] = TagDefs["ITALIC"] = TagDefs["EM"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = false; Tag.Is_OpenTag = true;
            Tag.Handler = &Transceiver::HandleUnderlineTag;
         TagDefs["U"] = TagDefs["UNDERLINE"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = false; Tag.Is_OpenTag = true;
            Tag.Handler = &Transceiver::HandleStrikeoutTag;
         TagDefs["S"] = TagDefs["STRIKE"] = TagDefs["STRIKEOUT"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = false; Tag.Is_OpenTag = true;
            DefAt(Tag, "FORE", "");
            DefAt(Tag, "BACK", "");
            Tag.Handler = &Transceiver::HandleColorTag;
         TagDefs["C"] = TagDefs["COLOR"] = TagDefs["COLOUR"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = false; Tag.Is_OpenTag = true;
            Tag.Handler = &Transceiver::HandleBrightTag;
         TagDefs["H"] = TagDefs["HIGH"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = false; Tag.Is_OpenTag = true;
            DefAt(Tag, "FACE", "");
            DefAt(Tag, "SIZE", "");
            DefAt(Tag, "COLOR", "");
            DefAt(Tag, "BACK", "");
            Tag.Handler = &Transceiver::HandleFontTag;
         TagDefs["FONT"] = Tag;}

        /* TEXT FORMATTING TAGS (SECURE) */
        /*  // -- disabled for now because they're not implemented
        {TagDef Tag; Tag.Is_EmptyTag = false;
            Tag.Handler = &Transceiver::HandleNobrTag;
         TagDefs["NOBR"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = false;
            Tag.Handler = &Transceiver::HandleParaTag;
         TagDefs["P"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = true;
            Tag.Handler = &Transceiver::HandleBRTag;
         TagDefs["BR"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = true;
            Tag.Handler = &Transceiver::HandleSBRTag;
         TagDefs["SBR"] = Tag;}
        */
        
        /* SYSTEM TAGS (SECURE) */
        {TagDef Tag; Tag.Is_EmptyTag = true;
            Tag.Handler = &Transceiver::HandleSupportTag;
         TagDefs["SUPPORT"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = true;
            Tag.Handler = &Transceiver::HandleVersionTag;
         TagDefs["VERSION"] = Tag;}
         
        #undef DefAt
    }

private:
    bool TranslateZcommand(const std::string& param)
    {
        #define NlMode(n) ModeMode=UntilNl,CurrentMode=n##Mode
        #define LkMode(n) ModeMode=ModeLocked,CurrentMode=n##Mode
        #define TgMode(n) ModeMode=UntilTagEnd,CurrentMode=n##Mode
        
        if(param == "2;0") { ClosingTagsMode=false; return true; }
        if(param == "2;1") { ClosingTagsMode=true; return true; }
        
        unsigned val = std::atoi(param.c_str());
        switch(val)
        {
            case 0: NlMode(Open); return true;
            case 1: NlMode(Secure); return true;
            case 2: NlMode(Locked); return true;
            case 3: Formatting.Reset(); NlMode(Open); return true;
            case 4: TgMode(Secure); return true;
            case 5: LkMode(Open); return true;
            case 6: LkMode(Secure); return true;
            case 7: LkMode(Locked); return true;
        }
        if(val >= 20 && val <= 99)
        {
            for(std::map<std::string/*tagname*/, TagDef>::const_iterator
                i = TagDefs.begin(); i != TagDefs.end(); ++i)
            {
                if(i->second.Tag == val)
                {
                    //fprintf(stderr, "Found TAG '%s' with value '%u'\n", i->first.c_str(), val);
                    Tag dummy;
                    dummy.IsClosingTag = false;
                    (this->*(i->second.Handler))(i->second, dummy);
                    LineEndCloseTags.push_back(i->first);
                    return true;
                }
            }
        }
        return false;
        
        #undef NlMode
        #undef LkMode
        #undef TgMode
    }
    
    bool TranslateTag(const Tag& tag)
    {
        /*
        fprintf(stderr, "Handling tag: '%s%s'%s\n",
            tag.IsClosingTag ? "/" : "",
            tag.TagName.c_str(),
            ClosingTagsMode ? " (automatically closed)" : ""
        );*/
        
        std::map<std::string/*tagname*/, TagDef>::const_iterator
            i = TagDefs.find(tag.TagName);
        if(i != TagDefs.end())
        {
            if(i->second.Is_EmptyTag && tag.IsClosingTag) return false;
            
            (this->*(i->second.Handler))(i->second, tag);
            
            return true;
        }
        return false;
    }
    
    void DoMacroExpansion(const std::string& macro)
    {
        //fprintf(stderr, "\nAdding priority input: '%s'\n", macro.c_str());
        PendingMacro += macro;
    }
    void DoFeedback(const std::string& feedback)
    {
        fprintf(stderr, "\nSending feedback: '%s'\n", feedback.c_str());
        PendingFeedback += feedback;
    }
    
    void RedoFormatting()
    {
        PendingResult += Formatting.GetUpdatedCodes();
    }

    void HandleChar(const unsigned char input_ch)
    {
        scrBegin;
        #define NextCh() scrReturnV
        for(;;)
        {
            //fprintf(stderr, "received ch: '%c'\n", input_ch);
            
Rehandle:
            if(input_ch == IAC) // telnet escape
            {
                static std::string telnetesc; telnetesc = "";
                telnetesc += input_ch;
                
                NextCh();
                if(input_ch == WILL)
                {
                    telnetesc += input_ch;
                    NextCh();
                    if(input_ch == 91) // MXP
                    {
                        PendingResult += (char)IAC;
                        PendingResult += (char)DO;
                        PendingResult += (char)91;
                        continue;
                    }
                }
                PendingResult += telnetesc;
                goto Literal;
            }
            if(CurrentMode != LockedMode)
            {
                if(input_ch == '\33')
                {
                    NextCh();
                    if(input_ch != '[') { PendingResult += '\33'; goto Rehandle; }
                    static std::string params; params="";
                NextParamCh:
                    NextCh();
                    if((input_ch >= '0' && input_ch <= '9') || input_ch==';')
                        { params += input_ch; goto NextParamCh; }
                    if(input_ch == 'z')
                        { if(!TranslateZcommand(params))
                            goto LiteralEscCmd;
                        }
                    else
                    { LiteralEscCmd:
                        PendingResult += "\33[";
                        PendingResult += params;
                        goto Rehandle;
                    }
                }
                else if(input_ch == '<') // tag
                {
                    static std::string tagbuf; tagbuf = "";
                    #define TagGetCh() { tagbuf += input_ch; NextCh(); }
                    static std::string tagname; tagname = "";
                    static bool is_closing; is_closing=false;
                    static std::string attrlist; attrlist="";
                    for(;;)
                    {
                        TagGetCh();
                        //fprintf(stderr, "Tagname processing: '%c'\n", input_ch);
                        if(tagname.empty() && input_ch == '/') { is_closing=true; continue; }
                        if(std::isspace(input_ch) || input_ch == '>' || input_ch == '/')
                        {
                            if(tagname.empty()) continue;
                            break;
                        }
                        if(tagname.empty() && input_ch == '>'
                        || tagname.size() > 48
                        || !is_valid_tagname_char(input_ch))
                            { InvalidTag: PendingResult += tagbuf; goto Rehandle; }
                        tagname += input_ch;
                    }
                    if(TagDefs.find(UCase(tagname)) == TagDefs.end()) goto InvalidTag;
                    
                    while(std::isspace(input_ch)) { TagGetCh(); }
                    static char quotemode; quotemode=0;
                    for(;;)
                    {
                        //fprintf(stderr, "Tagattr processing: '%c'\n", input_ch);
                        if(!quotemode && input_ch == '/') { TagGetCh(); continue; }
                        if(!quotemode && input_ch == '>') break;
                        attrlist += input_ch;
                        if(quotemode && input_ch==quotemode) { quotemode=false; }
                        else if(!quotemode)
                        {
                            if(input_ch=='"' || input_ch=='\'')
                            {
                                char prev = 0;
                                if(attrlist.size() > 1) prev = attrlist[attrlist.size()-2];
                                if(!prev || prev=='=' || std::isspace(prev))
                                    quotemode=input_ch;
                                else goto InvalidTag;
                            }
                        }
                        else if(!quotemode && !is_valid_unquoted_tagattr_char(input_ch))
                            goto InvalidTag;
                        else {}
                        TagGetCh();
                    }
                    
                    //fprintf(stderr, "name(%s)attr(%s)\n",
                    //    tagname.c_str(), attrlist.c_str());
                    
                    Tag tag;
                    tag.TagName      = UCase(tagname);
                    tag.IsClosingTag = is_closing;
                    if(ClosingTagsMode) tag.IsClosingTag = true; // force closing
                    tag.Params = ParseAttrList(attrlist, false);
                    if(!TranslateTag(tag)) goto InvalidTag;
                    #undef TagGetCh
                }
                else if(input_ch == '&') // entity
                {
                    static std::string entity; entity = "";
                    NextCh();
                    if(input_ch == '#')
                    {
                        static unsigned charvalue; charvalue = 0;
                        static unsigned base; base = 10;
                        entity += '#';
                        NextCh(); if(input_ch == 'x' || input_ch == 'X')
                                   { entity += input_ch; base=16; NextCh(); }
                        for(;;)
                        {
                            #define chr(min,max, sub,add) \
                                if(input_ch >= min && input_ch <= max) \
                                { charvalue = charvalue*base + (input_ch - sub + add); \
                                  entity += input_ch; \
                                  NextCh(); continue; }

                            chr('0','9', '0',0);
                            if(base == 16)
                            { chr('A','F', 'A',10);
                              chr('a','f', 'a',10); }
                            if(input_ch == ';') break;
                        FailedEntity:
                            PendingResult += "&";
                            PendingResult += entity;
                            goto Rehandle;
                        }
                        PendingResult += (char)charvalue;
                    }
                    else
                    {
                        while(input_ch != ';' && entity.size() < 16)
                            { entity += input_ch; NextCh(); }
                        if(input_ch != ';') goto FailedEntity;
                        std::map<std::string, EntityDef>::const_iterator
                            i = EntityDefs.find(entity);
                        if(i == EntityDefs.end()) goto FailedEntity;
                        
                        if(i->second.Is_Recursive)
                            DoMacroExpansion(i->second.Value);
                        else
                            PendingResult += i->second.Value;
                    }
                }
                else goto Literal;
            }
            else
            {
            Literal:
                PendingResult += input_ch;
                if(ModeMode == UntilTagEnd) goto ResetDefaultMode;
                if(input_ch == '\n' && ModeMode == UntilNl)
                {
                ResetDefaultMode:
                    CurrentMode = DefaultMode;
                    ModeMode    = ModeLocked;
                }
                if(input_ch == '\n')
                {
                    for(unsigned a=LineEndCloseTags.size(); a-->0; )
                    {
                        Tag dummy;
                        dummy.TagName      = LineEndCloseTags[a];
                        dummy.IsClosingTag = true;
                        TranslateTag(dummy);
                    }
                    LineEndCloseTags.clear();
                    ClosingTagsMode = false;
                }
            }
            NextCh();
        }
        scrFinishV;
    }
public:
    void ProcessBuf(const std::string& buf)
    {
        for(unsigned a=0; a<buf.size(); ++a)
        {
            HandleChar(buf[a]);
            
            while(!PendingMacro.empty())
            {
                const std::string macro = PendingMacro;
                PendingMacro = "";
                for(unsigned b=0; b<macro.size(); ++b)
                    HandleChar(macro[b]);
            }
        }
    }
public:
    Transceiver() :
        CurrentMode(DefaultMode),
        ModeMode(ModeLocked),
        ClosingTagsMode(false)
    {
        InitDefaultTags();
    }
    
    ssize_t Handle(ssize_t(*orig_recv)(int,void*,size_t,int),
                   int s,void*buf,size_t len,int flags)
    {
    Redo:
        if(!PendingFeedback.empty())
        {
            ssize_t took = send(s, &PendingFeedback[0], PendingFeedback.size(), 0);
            if(took > 0)
                PendingFeedback.erase(0, took);
        }
        
        if(len > 0 && !PendingResult.empty())
        {
            ssize_t take = std::min(len, PendingResult.size());
            std::memcpy(buf, PendingResult.data(), take);
            PendingResult.erase(0, take);
            return take;
        }
    
        ssize_t res = orig_recv(s,buf,len,flags);
        if(res <= 0) return res;

        const char* charbuf = (const char*)buf;
        std::string Buf( charbuf, charbuf + res);
        //PendingResult += Buf;
        ProcessBuf(Buf);
        goto Redo;
    }
    
    const std::string& GetResult() const { return PendingResult; }
};

extern "C" {

ssize_t recv(int s, void *buf, size_t len, int flags)
{
    /* Find the real recv function */
    typedef ssize_t (*old_t)(int,void*,size_t,int);
    static old_t old = (old_t)caller("recv");
    
    static class Transceiver Transceiver;
    return Transceiver.Handle(old, s,buf,len,flags);
}

}

/*
int main(void)
{
    // Testing if everything works as specified.
    
    Transceiver t;
    t.ProcessBuf("<!ELEMENT Auction Tag=20>"
                 "<C #C08040> aiueo </C> x\n"
                 "<!TAG 20 Fore=Red>"
                 "\33[20zEKA\n"
                 "<Auction>Toka</Auction>TokaEnd\n"
                 "<!TAG 20 Fore=Blue>"
                 "\33[20zKOLMAS\n"
                 "<Auction>Neljs</Auction>NeljsEnd\n"
                 );
    printf("Got: '%s'\n", t.GetResult().c_str());
}*/
